import polars as pl
import numpy as np
import pandas as pd
from itertools import product
from scipy.spatial import distance
from scipy.spatial.distance import pdist, squareform, cdist
from scipy import sparse

################################################################################

def euclidean_dist_matrix(X):
    """
    Calculates the Euclidean distance matrix for a data matrix using SciPy.

    Parameters (inputs)
    ----------
    X: a Pandas or Polars DataFrame or a NumPy array. It represents a data matrix.

    Returns (outputs)
    -------
    M: the Euclidean distance matrix between the rows of `X`.
    """
    
    # Convert to NumPy array if input is a DataFrame.
    if isinstance(X, pl.DataFrame):
        X = X.to_numpy()
    if isinstance(X, pd.DataFrame):
        X = X.to_numpy()    
            
    # Compute the pairwise distances using pdist and convert to a square form.
    M = squareform(pdist(X, metric='euclidean'))
    
    return M

################################################################################

def euclidean_dist(xi, xr) :
    """
    Calculates the Euclidean distance between a pair of vectors.

    Parameters (inputs)
    ----------
    xi, xr: a pair of Pandas or Polars Series or DataFrames. 
            They represent a couple of statistical observations of quantitative variables. 

    Returns (outputs)
    -------
    The Euclidean distance between the observations `xi` and `xr`.
    """

    if isinstance(xi, (pl.DataFrame, pd.DataFrame)) :
        xi = xi.to_numpy().flatten()
    elif isinstance(xi, (pd.Series, pl.Series)) :
        xi = xi.to_numpy() 
    if isinstance(xr, (pl.DataFrame, pd.DataFrame)) :
        xr = xr.to_numpy().flatten()
    elif isinstance(xr, (pd.Series, pl.Series)) :
        xr = xr.to_numpy() 

    return distance.euclidean(xi, xr)

################################################################################

def minkowski_dist_matrix(X, q):
    """
    Calculates the Minkowski distance matrix for a data matrix using SciPy.

    Parameters (inputs)
    ----------
    X: a Pandas or Polars DataFrame or a NumPy array. It represents a data matrix.
    q: the parameters that defines the Minkowski form. Some particular cases: q=1 := Manhattan, q=2 := Euclidean.

    Returns (outputs)
    -------
    M: the Minkowski(`q`) distance matrix between the rows of `X`.
    """   
    
    if isinstance(X, pl.DataFrame):
        X = X.to_numpy()
    if isinstance(X, pd.DataFrame):
        X = X.to_numpy()    
            
    # Compute the pairwise distances using pdist and convert to a square form.
    M = squareform(pdist(X, metric='minkowski', p=q))
    
    return M

################################################################################

def minkowski_dist(xi, xr, q) :
    """
    Calculates the Minkowski distance between a pair of vectors.

    Parameters (inputs)
    ----------
    xi, xr: a pair of quantitative vectors. They represent a couple of statistical observations.
    q: the parameters that defines the Minkowski form. Some particular cases: q=1 := Manhattan, q=2 := Euclidean.

    Returns (outputs)
    -------
    The Minkowki(`q`) distance between the observations `xi` and `xr`.
    """

    if isinstance(xi, (pl.DataFrame, pd.DataFrame)) :
        xi = xi.to_numpy().flatten()
    elif isinstance(xi, (pd.Series, pl.Series)) :
        xi = xi.to_numpy() 
    if isinstance(xr, (pl.DataFrame, pd.DataFrame)) :
        xr = xr.to_numpy().flatten()
    elif isinstance(xr, (pd.Series, pl.Series)) :
        xr = xr.to_numpy() 

    return distance.minkowski(xi, xr, q)

################################################################################

def canberra_dist_matrix(X):
    """
    Calculates the Canberra distance matrix for a data matrix using SciPy.

    Parameters (inputs)
    ----------
    X: a pandas/polars DataFrame or a NumPy array. It represents a data matrix.

    Returns (outputs)
    -------
    M: the Canberra distance matrix between the rows of `X`.
    """

    if isinstance(X, pl.DataFrame):
        X = X.to_numpy()
    if isinstance(X, pd.DataFrame):
        X = X.to_numpy()    
            
    # Compute the pairwise distances using pdist and convert to a square form.
    M = squareform(pdist(X, metric='canberra'))
    
    return M

################################################################################

def canberra_dist(xi, xr) :
    """
    Calculates the Canberra distance between a pair of vectors.

    Parameters (inputs)
    ----------
    xi, xr: a pair of quantitative vectors. They represent a couple of statistical observations.

    Returns (outputs)
    -------
    The Canberra distance between the observations `xi` and `xr`.
    """

    if isinstance(xi, (pl.DataFrame, pd.DataFrame)) :
        xi = xi.to_numpy().flatten()
    elif isinstance(xi, (pd.Series, pl.Series)) :
        xi = xi.to_numpy() 
    if isinstance(xr, (pl.DataFrame, pd.DataFrame)) :
        xr = xr.to_numpy().flatten()
    elif isinstance(xr, (pd.Series, pl.Series)) :
        xr = xr.to_numpy() 

    return distance.canberra(xi, xr)

################################################################################

def pearson_dist_matrix(X):
    """
    Calculates the Pearson distance matrix for a data matrix using SciPy.

    Parameters (inputs)
    ----------
    X: a pandas/polars DataFrame or a NumPy array. It represents a data matrix.

    Returns (outputs)
    -------
    M: the Pearson distance matrix between the rows of X.
    """

    if isinstance(X, pl.DataFrame):
        X = X.to_numpy()
    if isinstance(X, pd.DataFrame):
        X = X.to_numpy()    
        
    # Compute the pairwise distances using pdist and convert to a square form.
    M = squareform(pdist(X, metric='seuclidean'))
    
    return M

################################################################################

def mahalanobis_dist_matrix(X):
    """
    Calculates the classical Mahalanobis distance matrix for a data matrix `X`.

    Parameters
    ----------
    X : pandas.DataFrame, polars.DataFrame, or np.ndarray
        Data matrix of shape (n_samples, n_features).

    Returns
    -------
    D : np.ndarray
        Symmetric matrix (n_samples x n_samples) of Mahalanobis distances.
    """
    
    # Convert to numpy array if needed
    if isinstance(X, pl.DataFrame):
        X = X.to_numpy()
    elif isinstance(X, pd.DataFrame):
        X = X.to_numpy()

    # Center the data
    X_centered = X - np.mean(X, axis=0)

    # Classical covariance matrix
    S = np.cov(X_centered, rowvar=False)

    # Use pseudo-inverse for numerical stability
    S_pinv = np.linalg.pinv(S)

    # Symmetrize just in case
    S_pinv = (S_pinv + S_pinv.T) / 2

    # Compute Mahalanobis distance matrix
    D = cdist(X_centered, X_centered, metric='mahalanobis', VI=S_pinv)

    return D

################################################################################

def mahalanobis_dist(xi, xr, S) :
    """
    Calculates the Mahalanobis distance between a pair of vectors.

    Parameters (inputs)
    ----------
    xi, xr: a pair of quantitative vectors. They represent a couple of statistical observations.
    S: the covariance matrix of the data matrix to which `xi` and `xr` belong.

    Returns (outputs)
    -------
    The Mahalanobis distance between the observations `xi` and `xr`.
    """

    if isinstance(xi, (pl.DataFrame, pd.DataFrame)) :
        xi = xi.to_numpy().flatten()
    elif isinstance(xi, (pd.Series, pl.Series)) :
        xi = xi.to_numpy() 
    if isinstance(xr, (pl.DataFrame, pd.DataFrame)) :
        xr = xr.to_numpy().flatten()
    elif isinstance(xr, (pd.Series, pl.Series)) :
        xr = xr.to_numpy() 
    
    S_inv = np.linalg.inv(S) 
    dist = distance.mahalanobis(xi, xr, S_inv)

    return dist

################################################################################

def mad(Xj) :
    """
    Compute the median absolute deviation of a statistical variable.

    Parameters (inputs)
    ----------
    Xj: a vector representing a quantitative statistical variable.  

    Returns (outputs)
    -------
    MAD: median absolute deviation of `Xj`.
    """

    if isinstance(Xj, pl.Series):
        Xj = Xj.to_numpy()
    elif isinstance(Xj, pd.Series):
        Xj = Xj.to_numpy()

    mad_ = np.median(np.abs(Xj - np.median(Xj)))

    return mad_

################################################################################

def Xj_trimmed(Xj, alpha) : 
    """
    Compute the trimmed version of a statistical variable.

    Parameters (inputs)
    ----------
    Xj : a vector representing a quantitative statistical variable.  
    alpha : a real number in [0,1] that defines the trimming level. 

    Returns (outputs)
    -------
    result: the `alpha` trimmed version of `Xj`.
    """

    if isinstance(Xj, pl.Series):
        Xj = Xj.to_numpy()
    elif isinstance(Xj, pd.Series):
        Xj = Xj.to_numpy()

    lower_bound = np.quantile(Xj, q=alpha/2)  
    upper_bound = np.quantile(Xj, q=1-alpha/2) 
    mask = np.logical_and(Xj >= lower_bound, Xj <= upper_bound)
    Xj_trimmed_ = Xj[mask]
    
    return Xj_trimmed_

################################################################################

def Xj_winsorized(Xj, alpha):
    """
    Compute the winsorized version of a quantitative variable.

    Parameters
    ----------
    Xj : a vector representing a quantitative statistical variable.  
    alpha : a real number in [0,1] that defines the winsorizing level. 

    Returns
    -------
    result: the `alpha` winsorized version of Xj.
    """

    if isinstance(Xj, pl.Series):
        Xj = Xj.to_numpy()
    elif isinstance(Xj, pd.Series):
        Xj = Xj.to_numpy()

    # If Xj is a vector of zeros, return Xj.
    if np.all(Xj == 0):
        return Xj

    lower_bound = np.quantile(Xj, q=alpha/2)
    upper_bound = np.quantile(Xj, q=1-alpha/2)

    # Clip the values: values smaller than lower_bound are set to lower_bound,
    #  those larger than upper_bound are set to upper_bound, 
    # and the ones in the middle are not transform.
    Xj_winsorized_ = np.clip(Xj, lower_bound, upper_bound)

    return Xj_winsorized_

################################################################################

def robust_var(Xj, method, alpha=None) :
    """
    Compute the robust variance of `Xj` allowing different methods.

    Parameters
    ----------
    Xj : a vector representing a quantitative statistical variable.  
    method: the method to be used for computing the robust variance of `Xj`. Must be an string in ['MAD', 'trimmed', 'winsorized'].
    alpha : a real number in [0,1] that is used if `method` is 'trimmed' or 'winsorized'. 

    Returns
    -------
    result: the robust variance of `Xj` computed by the method `method`.
    """

    if method == 'MAD' :
        
        return mad(Xj)**2

    if method == 'trimmed' :
        
        return np.var(Xj_trimmed(Xj, alpha))

    if method == 'winsorized' :
        
        return np.var(Xj_winsorized(Xj, alpha))
    
################################################################################

def robust_corr(Xj, Xr, method, alpha=None) :
    """
    Compute the robust correlation between `Xj` and `Xr` by different methods.

    Parameters
    ----------
    Xj, Xr : two vectors representing a quantitative statistical variables.  
    method: the method to be used for computing the robust variance of `Xj`. Must be an string in ['MAD', 'trimmed', 'winsorized'].
    alpha : a real number in [0,1] that is used if `method` is 'trimmed' or 'winsorized'. 

    Returns
    -------
    result: the robust correlation between `Xj` and `Xr` computed by the method `method`.
    """

    if isinstance(Xj, pl.Series):
        Xj = Xj.to_numpy()
    elif isinstance(Xj, pd.Series):
        Xj = Xj.to_numpy()
    if isinstance(Xr, pl.Series):
        Xr = Xr.to_numpy()
    elif isinstance(Xr, pd.Series):
        Xr = Xr.to_numpy()

    # Si la varianza robusta de X_j es cero, la version estandarizada de X_j es la propia X_j.
    if robust_var(Xj, method, alpha) == 0 :
        Xj_std = Xj
    # Si la varianza robusta de X_j es distinta de cero.
    else :
       # Se estandariza X_j como se especifica en la seccion 7.2.2.
        Xj_std = Xj / np.sqrt(robust_var(Xj, method, alpha))
    # Si la varianza robusta de X_r es cero, la version estandarizada de X_r es la propia X_r.
    if np.sqrt(robust_var(Xr, method, alpha)) == 0 :
        Xr_std = Xr
    # Si la varianza robusta de X_res distinta de cero.
    else :
      # Se estandariza X_r como se especifica en la seccion 7.2.2.  
      Xr_std = Xr / np.sqrt(robust_var(Xr, method, alpha))
      
    # Se calcula la correlacion robusta como se especifica en la seccion 7.2.2, evitando problemas de divisionalidad.
    robust_var_3 = robust_var(Xj_std + Xr_std, method, alpha)
    robust_var_4 = robust_var(Xj_std - Xr_std, method, alpha)
    if (robust_var_3 + robust_var_4) == 0 :
        robust_corr = (robust_var_3 - robust_var_4) 
    else : 
        robust_corr = (robust_var_3 - robust_var_4) / (robust_var_3 + robust_var_4)
    return robust_corr

################################################################################

def R_robust(X, method, alpha=None) :
    """
    Computes the robust correlation matrix of a given data matrix `X`.

    Parameters
    ----------
    X : a pandas/polars data-frame or a numpy array. 
    method : the method used to compute the robust correlation matrix.  Must be an string in ['MAD', 'trimmed', 'winsorized']. 
    alpha : a real number in [0,1] that is used if `method` is 'trimmed' or 'winsorized'.

    Returns
    -------
    M : the robust correlation matrix for `X`.
    """

    if isinstance(X, pl.DataFrame):
        X = X.to_numpy()
    elif isinstance(X, pd.DataFrame):
        X = X.to_numpy()
    
    p = X.shape[1]
    M =  np.zeros((p,p))

    for j,r in product(range(0,p), range(0,p)) :

        M[j,r] = robust_corr(Xj=X[:,j], Xr=X[:,r], method=method, alpha=alpha)

    return M

################################################################################

def delvin_trans(M, epsilon=0.05) : 
    """
    Applies the Delvin transformation for the matrix `M` passed as input to make it positive definite or closer to it.

    Parameters (inputs)
    ----------
    M : a pandas/polars data-frame or a numpy array. 
    epsilon : parameter involved in the Delvin transformation that must be a close to zero positive number. epsilon=0.05 is recommended.

    Returns (outputs)
    -------
    M : the Delvin transformation of the input matrix `M`.
    """

    if isinstance(M, pl.DataFrame):
        M = M.to_numpy()
    elif isinstance(M, pd.DataFrame):
        M = M.to_numpy()   

    # Se define la funcion z.
    def z(x) : 
       return np.arctanh(x)
    
    # Se define la funcion z^{-1}
    def z_inv(x) :  
       # La arctanh es la inversa de tanh, por tanto, la inversa de arctanh es tanh.
       return np.tanh(x) 
    
    # Se define la funcion g.
    def g(i,j, M) :
        if i == j :             
            return 1
        else:
            if np.abs(M[i,j]) <= z(epsilon) :  
                    return 0
            elif M[i,j] < - z(epsilon) : 
                    return z_inv(M[i,j] + epsilon)
            elif M[i,j] > z(epsilon) : 
                    return z_inv(M[i,j] - epsilon)
    
    # Se crea una matriz cuyos elementos son el resultado de aplicar la funcion g sobre matrix elemento a elemento.
    p = M.shape[1]
    M_new =  np.zeros((p,p))
    
    for i,j in product(range(0,p), range(0,p)) :
            
            M_new[i,j] = g(i,j, M)

    return M_new  

################################################################################

def delvin_algorithm(M, epsilon, n_iters):
    """
    Applies the Delvin algorithm on the matrix `M` passed as input 
    to make it positive definite by applying on it the Delvin transformation as many iterations as needed.

    Parameters (inputs)
    ----------
    M: a pandas/polars data-frame or a numpy array. 
    epsilon : parameter used by the Delvin transformation. epsilon=0.05 is recommended.
    n_iter : maximum number of iterations run by the algorithm.

    Returns (outputs)
    -------
    M_new : the resulting matrix of applying the Delvin algorithm on `M`.
    """

    M_new = M.copy()
    # Se inicializa i=0 para entrar en el bucle while
    i = 0
    # Mientras i sea inferior o igual a n_iter, el bucle continua ejecutandose.
    while i < n_iters:
        # Si new_matrix ya es definida positiva (todos sus autovalores son positivos), se devuelve new_matrix. 
        # En otro caso, se le aplica la transformacion de Delvin y se vuelve a comprobar si es definida positiva.
        if np.all(np.linalg.eigvals(M_new) > 0):
            return M_new, i
        else:
            M_new = delvin_trans(M=M_new, epsilon=epsilon)
            i = i + 1
    
    return M_new, i

################################################################################

def S_robust(X, method, epsilon, n_iters, alpha=None, weights=None):
    """
    Computes the robust covariance of the data matrix `X` by different methods.

    Parameters (inputs)
    ----------
    X: a pandas/polars data-frame or a numpy array. 
    method: the method to be used to compute the robust covariance. Must be an string in ['MAD', 'trimmed', 'winsorized']. 
    alpha : a real number in [0,1] that is used if `method` is 'trimmed' or 'winsorized'.
    epsilon : parameter used by the Delvin transformation. epsilon=0.05 is recommended.
    n_iter : maximum number of iterations run by the Delvin algorithm.
    weights: the sample weights. Only used if provided.

    Returns (outputs)
    -------
    S_robust : the robust covariance matrix computed for `X`.
    """

    if isinstance(X, pl.DataFrame):
        X = X.to_numpy()
    elif isinstance(X, pd.DataFrame):
        X = X.to_numpy() 

    if weights is None:
        # Se calcula la matriz de correlaciones robustas para Data.
        R_robust_ = R_robust(X, method, alpha)
        # Se aplica el algoritmo de Delvin a la matriz de correlaciones robustas calculada.
        R_robust_, i = delvin_algorithm(M=R_robust_, epsilon=epsilon, n_iters=n_iters)
        # Se calcula la matriz de covarianzas robustas a partir de la matriz de correlaciones robustas.
        S_robust = np.diag(np.std(X, axis=0)) @ R_robust_ @ np.diag(np.std(X, axis=0)) 

    else:
        w = weights
        n = len(X)
        Dw = sparse.diags(w)
        I = np.identity(n)
        ones_arr = np.ones(n)
        Jw = np.sqrt(Dw) @ (I - ones_arr @ w.T)
        Xw = Jw @ X # Computational problems when n is too large since is an n x n matrix.
        # Se calcula la matriz de correlaciones robustas para Data.
        R_robust_ = R_robust(Xw, method, alpha)
        # Se aplica el algoritmo de Delvin a la matriz de correlaciones robustas calculada.
        R_robust_, i = delvin_algorithm(M=R_robust_, epsilon=epsilon, n_iters=n_iters)
        # Se calcula la matriz de covarianzas robustas a partir de la matriz de correlaciones robustas.
        S_robust = np.diag(np.std(Xw, axis=0)) @ R_robust_ @ np.diag(np.std(Xw, axis=0)) 

    return S_robust

################################################################################

def robust_maha_dist_matrix(X, S_robust):
    """
    Calculates the Robust Mahalanobis distance matrix for a data matrix `X`
    using a robust estimation of the covariance matrix.

    Parameters
    ----------
    X : pandas.DataFrame, polars.DataFrame, or np.ndarray
        The input data matrix with shape (n_samples, n_features).
    
    S_robust : np.ndarray
        Robust covariance matrix (e.g., from MCD or a trimmed estimator).
        Should be of shape (n_features, n_features).

    Returns
    -------
    D : np.ndarray
        Symmetric matrix (n_samples, n_samples) of Mahalanobis distances.
    """

    # Convert input to NumPy array if needed
    if isinstance(X, pl.DataFrame):
        X = X.to_numpy()
    elif isinstance(X, pd.DataFrame):
        X = X.to_numpy()

    # Center the data (important for Mahalanobis)
    X_centered = X - np.mean(X, axis=0)

    # Use pseudo-inverse for stability
    S_robust_pinv = np.linalg.pinv(S_robust)

    # Force symmetry (sometimes lost by numerical error)
    S_robust_pinv = (S_robust_pinv + S_robust_pinv.T) / 2

    # Compute pairwise Mahalanobis distances with cdist
    D = cdist(X_centered, X_centered, metric='mahalanobis', VI=S_robust_pinv)

    return D

################################################################################

def robust_maha_dist(xi, xr, S_robust) :
    """
    Calculates the Robust Mahalanobis distance between a pair of vectors.

    Parameters (inputs)
    ----------
    xi, xr: a pair of quantitative vectors. They represent a couple of statistical observations.
    S_robust: the robust covariance matrix of the data matrix to which `xi` and `xr` belong.

    Returns (outputs)
    -------
    The Robust Mahalanobis distance between the observations `xi` and `xr`.
    """

    if isinstance(xi, (pl.DataFrame, pd.DataFrame)) :
        xi = xi.to_numpy().flatten()
    elif isinstance(xi, (pd.Series, pl.Series)) :
        xi = xi.to_numpy() 
    if isinstance(xr, (pl.DataFrame, pd.DataFrame)) :
        xr = xr.to_numpy().flatten()
    elif isinstance(xr, (pd.Series, pl.Series)) :
        xr = xr.to_numpy() 

    X = np.array([xi, xr])
    dist_xi_xr = robust_maha_dist_matrix(X, S_robust)
    dist_xi_xr = dist_xi_xr[0,1]
    
    return dist_xi_xr

################################################################################


import polars as pl
import numpy as np
import pandas as pd
from robust_mixed_dist.quantitative import (euclidean_dist_matrix, euclidean_dist, minkowski_dist_matrix, 
                                      minkowski_dist, canberra_dist_matrix, canberra_dist, pearson_dist_matrix, 
                                      mahalanobis_dist_matrix, mahalanobis_dist, robust_maha_dist_matrix, robust_maha_dist, S_robust)
from robust_mixed_dist.binary import sokal_dist_matrix, sokal_dist, jaccard_dist_matrix, jaccard_dist
from robust_mixed_dist.multiclass import hamming_dist_matrix, hamming_dist

################################################################################

def get_dist_matrix_functions():
        
    dist_matrix = {}
    dist_matrix['euclidean'] = euclidean_dist_matrix
    dist_matrix['minkowski'] = minkowski_dist_matrix
    dist_matrix['canberra'] = canberra_dist_matrix
    dist_matrix['pearson'] = pearson_dist_matrix
    dist_matrix['mahalanobis'] = mahalanobis_dist_matrix
    dist_matrix['robust_mahalanobis'] = robust_maha_dist_matrix
    dist_matrix['sokal'] = sokal_dist_matrix
    dist_matrix['jaccard'] = jaccard_dist_matrix
    dist_matrix['hamming'] = hamming_dist_matrix

    return dist_matrix

################################################################################

def get_dist_functions():

    dist = {}
    dist['euclidean'] = euclidean_dist
    dist['minkowski'] = minkowski_dist
    dist['canberra'] = canberra_dist
    dist['mahalanobis'] = mahalanobis_dist
    dist['robust_mahalanobis'] = robust_maha_dist
    dist['sokal'] = sokal_dist
    dist['jaccard'] = jaccard_dist
    dist['hamming'] = hamming_dist

    return dist

################################################################################

def vg(D_2):
    """
    Calculates the geometric variability of the squared distance matrix `D_2` passed as input.

    Parameters (inputs)
    ----------
    D_2: a numpy array. It should represent an squared distance matrix.

    Returns (outputs)
    -------
    VG: the geometric variability of the squared distance matrix `D_2`.
    """
    n = len(D_2)
    VG = (1/(2*(n**2)))*np.sum(D_2)
    # TO DO: version managing weights
    return VG

################################################################################

def get_dist_matrices(X, p1, p2, p3, d1='euclidean', d2='sokal', d3='matching', q=1, 
                      robust_method='trimmed', epsilon=0.05, alpha=0.05, n_iters=20, weights=None):
    """
    Calculates the distance matrices that are involved in the Generalized Gower distance.
            
    Parameters:
      X: a pandas/polars data-frame or a numpy array. Represents a data matrix.
      p1, p2, p3: number of quantitative, binary and multi-class variables in the considered data matrix, respectively. Must be a non negative integer.
      d1: name of the distance to be computed for quantitative variables. Must be an string in ['euclidean', 'minkowski', 'canberra', 'mahalanobis', 'robust_mahalanobis']. 
      d2: name of the distance to be computed for binary variables. Must be an string in ['sokal', 'jaccard'].
      d3: name of the distance to be computed for multi-class variables. Must be an string in ['matching'].
      q: the parameter that defines the Minkowski distance. Must be a positive integer.
      robust_method: the robust_method to be used for computing the robust covariance matrix. Only needed when d1 = 'robust_mahalanobis'.
      epsilon: parameter used by the Delvin algorithm that is used when computing the robust covariance matrix. Only needed when d1 = 'robust_mahalanobis'.
      n_iter: maximum number of iterations used by the Delvin algorithm. Only needed when d1 = 'robust_mahalanobis'.
      weights: the sample weights. Only used if provided and d1 = 'robust_mahalanobis'.  
        
    Returns:
      D1, D2, D3: the distances matrices associated to the quantitative, binary and multi-class variables, respectively.
    """
 
    if isinstance(X, (pl.DataFrame, pd.DataFrame)):
        X = X.to_numpy()

    dist_matrix = get_dist_matrix_functions()

    n = len(X)
    X_quant = X[:, 0:p1] 
    X_bin = X[:, (p1):(p1+p2)]
    X_multi = X[:, (p1+p2):(p1+p2+p3)]

    # Define D1 based on d1 and p1
    if p1 > 0:
        if d1 == 'minkowski':
            D1 = dist_matrix[d1](X_quant, q)
        elif d1 == 'robust_mahalanobis':
            S_robust_ = S_robust(X=X_quant, method=robust_method, alpha=alpha,
                                    epsilon=epsilon, n_iters=n_iters, 
                                    weights=weights)
            D1 = dist_matrix[d1](X_quant, S_robust=S_robust_)
        else:
            D1 = dist_matrix[d1](X_quant)
    elif p1 == 0:
        D1 = np.zeros((n, n))
    # Define D2 based on p2
    D2 = dist_matrix[d2](X_bin) if p2 > 0 else np.zeros((n, n)) 
    # Define D3 based on p3
    D3 = dist_matrix[d3](X_multi) if p3 > 0 else np.zeros((n, n))

    return D1, D2, D3

################################################################################

def get_distances(xi, xr, p1, p2, p3, d1='euclidean', d2='sokal', d3='matching', q=1, S=None, S_robust=None):
    """
    Calculates the distances between observations that are involved in the Generalized Gower distance.
       
    Parameters:
        xi, xr: a pair of quantitative vectors. They represent a couple of statistical observations.
        p1, p2, p3: number of quantitative, binary and multi-class variables in the considered data matrix, respectively. Must be a non negative integer.
        d1: name of the distance to be computed for quantitative variables. Must be an string in ['euclidean', 'minkowski', 'canberra', 'mahalanobis', 'robust_mahalanobis']. 
        d2: name of the distance to be computed for binary variables. Must be an string in ['sokal', 'jaccard'].
        d3: name of the distance to be computed for multi-class variables. Must be an string in ['matching'].
        q: the parameter that defines the Minkowski distance. Must be a positive integer.
        S: the covariance matrix of the considered data matrix.
        S_robust: the robust covariance matrix of the considered data matrix.
                  
    Returns:
        dist1, dist2, dist3: the distances values associated to the quantitative, binary and multi-class observations, respectively.
    """

    if isinstance(xi, (pl.DataFrame, pd.DataFrame)) :
        xi = xi.to_numpy().flatten()
    elif isinstance(xi, (pd.Series, pl.Series)) :
        xi = xi.to_numpy() 
    if isinstance(xr, (pl.DataFrame, pd.DataFrame)) :
        xr = xr.to_numpy().flatten()
    elif isinstance(xr, (pd.Series, pl.Series)) :
        xr = xr.to_numpy() 

    dist = get_dist_functions()
                   
    xi_quant = xi[0:p1] ; xr_quant = xr[0:p1] ; 
    xi_bin = xi[(p1):(p1+p2)] ; xr_bin = xr[(p1):(p1+p2)]
    xi_multi = xi[(p1+p2):(p1+p2+p3)] ; xr_multi = xr[(p1+p2):(p1+p2+p3)]

    if p1 > 0:
        if d1 == 'minkowski':
            dist1 = dist[d1](xi_quant, xr_quant, q=q)
        elif d1 == 'robust_mahalanobis':
            dist1 = dist[d1](xi_quant, xr_quant, S_robust=S_robust)
        elif d1 == 'mahalanobis':
            dist1 = dist[d1](xi_quant, xr_quant, S=S)
        else:
            dist1 = dist[d1](xi_quant, xr_quant)
    elif p1 == 0:
        dist1 = 0

    dist2 = dist[d2](xi_bin, xr_bin) if p2 > 0 else 0
    dist3 = dist[d3](xi_multi, xr_multi) if p3 > 0 else 0

    return dist1, dist2, dist3
    
################################################################################

def vg_ggower_estimation(X, p1, p2, p3, d1='euclidean', d2='sokal', d3='matching', 
                         q=1, robust_method='trimmed', epsilon=0.05, alpha=0.05, 
                         n_iters=20, weights=None): 
    """
    Calculates the geometric variability of an Generalized Gower distance matrix.

    Parameters:
        X: a pandas/polars data-frame or a numpy array. Represents a data matrix.
        p1, p2, p3: number of quantitative, binary and multi-class variables in the considered data matrix, respectively. Must be a non negative integer.
        d1: name of the distance to be computed for quantitative variables. Must be an string in ['euclidean', 'minkowski', 'canberra', 'mahalanobis', 'robust_mahalanobis']. 
        d2: name of the distance to be computed for binary variables. Must be an string in ['sokal', 'jaccard'].
        d3: name of the distance to be computed for multi-class variables. Must be an string in ['matching'].
        q: the parameter that defines the Minkowski distance. Must be a positive integer.
        robust_method: the robust_method to be used for computing the robust covariance matrix. Only needed when d1 = 'robust_mahalanobis'.
        epsilon: parameter used by the Delvin algorithm that is used when computing the robust covariance matrix. Only needed when d1 = 'robust_mahalanobis'.
        n_iter: maximum number of iterations used by the Delvin algorithm. Only needed when d1 = 'robust_mahalanobis'.
        weights: the sample weights. Only used if provided and d1 = 'robust_mahalanobis'.  
            
    Returns:
        VG1, VG2, VG3: the geometric variabilities of the distances matrices associated to the quantitative, binary and multi-class variables, respectively.
    """

    D1, D2, D3 = get_dist_matrices(X=X, p1=p1, p2=p2, p3=p3, d1=d1, d2=d2, d3=d3,
                                   q=q, robust_method=robust_method, epsilon=epsilon,
                                   alpha=alpha, n_iters=n_iters, weights=weights)
       
    D1_2, D2_2, D3_2 = D1**2, D2**2, D3**2
    VG1, VG2, VG3 = vg(D1_2), vg(D2_2), vg(D3_2)

    return VG1, VG2, VG3

################################################################################

def vg_ggower_fast_estimation(X, p1, p2, p3, d1='euclidean', d2='sokal', d3='matching',
                         robust_method='trimmed', epsilon=0.05, alpha=0.05, n_iters=20, q=1,
                         VG_sample_size=300, VG_n_samples=5, random_state=123, weights=None):
    """
    Calculates a fast estimation of the geometric variability of an squared Generalized Gower distance matrix.
            
    Parameters:
        X: a pandas/polars data-frame or a numpy array. Represents a data matrix.
        p1, p2, p3: number of quantitative, binary and multi-class variables in the considered data matrix, respectively. Must be a non negative integer.
        d1: name of the distance to be computed for quantitative variables. Must be an string in ['euclidean', 'minkowski', 'canberra', 'mahalanobis', 'robust_mahalanobis']. 
        d2: name of the distance to be computed for binary variables. Must be an string in ['sokal', 'jaccard'].
        d3: name of the distance to be computed for multi-class variables. Must be an string in ['matching'].
        q: the parameter that defines the Minkowski distance. Must be a positive integer.
        robust_method: the robust_method to be used for computing the robust covariance matrix. Only needed when d1 = 'robust_mahalanobis'.
        epsilon: parameter used by the Delvin algorithm that is used when computing the robust covariance matrix. Only needed when d1 = 'robust_mahalanobis'.
        n_iter: maximum number of iterations used by the Delvin algorithm. Only needed when d1 = 'robust_mahalanobis'.
        weights: the sample weights. Only used if provided and d1 = 'robust_mahalanobis'.  
        VG_sample_size: sample size to be used to make the estimation of the geometric variability.
        VG_n_samples: number of samples to be used to make the estimation of the geometric variability.
        random_state: the random seed used for the (random) sample elements.
        
    Returns:
        VG1, VG2, VG3: the geometric variabilities of the distances matrices associated to the quantitative, binary and multi-class variables, respectively.
    """
        
    if isinstance(X, (pl.DataFrame, pd.DataFrame)) :
        X = X.to_numpy()  
    
    n = len(X)
    VG1_list, VG2_list, VG3_list = [], [], []

    for i in range(0, VG_n_samples) :

        np.random.seed(random_state + i)
        index = np.arange(0,n)
        sample_index = np.random.choice(index, size=VG_sample_size)
        X_sample = X[sample_index,:].copy()
        
        if weights is not None:
            sample_weights = weights[sample_index].copy() 
        else:
            sample_weights = None
        
        VG1, VG2, VG3 = vg_ggower_estimation(X=X_sample, p1=p1, p2=p2, p3=p3, d1=d1, d2=d2, d3=d3, q=q,
                                            robust_method=robust_method, epsilon=epsilon, alpha=alpha, 
                                            n_iters=n_iters, weights=sample_weights)
        
        VG1_list.append(VG1) ; VG2_list.append(VG2) ; VG3_list.append(VG3) 

    VG1 = np.mean(VG1_list) ; VG2 = np.mean(VG2_list) ; VG3 = np.mean(VG3_list)

    return VG1, VG2, VG3

################################################################################
    
class GGowerDistMatrix: 
    """
    Calculates the Generalized Gower matrix for a data matrix.
    """

    def __init__(self, p1, p2, p3, d1='euclidean', d2='sokal', d3='matching', q=1, robust_method='trimmed', epsilon=0.05, alpha=0.05, n_iters=20,
                 fast_VG=False, VG_sample_size=300, VG_n_samples=5, random_state=123, weights=None):
        """
        Constructor method.
        
        Parameters:
            p1, p2, p3: number of quantitative, binary and multi-class variables in the considered data matrix, respectively. Must be a non negative integer.
            d1: name of the distance to be computed for quantitative variables. Must be an string in ['euclidean', 'minkowski', 'canberra', 'mahalanobis', 'robust_mahalanobis']. 
            d2: name of the distance to be computed for binary variables. Must be an string in ['sokal', 'jaccard'].
            d3: name of the distance to be computed for multi-class variables. Must be an string in ['hamming'].
            q: the parameter that defines the Minkowski distance. Must be a positive integer.
            metrobust_methodhod: the robust_method to be used for computing the robust covariance matrix. Only needed when d1 = 'robust_mahalanobis'.
            alpha : a real number in [0,1] that is used if `robust_method` is 'trimmed' or 'winsorized'. Only needed when d1 = 'robust_mahalanobis'.
            epsilon : parameter used by the Delvin transformation. epsilon=0.05 is recommended. Only needed when d1 = 'robust_mahalanobis'.
            n_iter : maximum number of iterations run by the Delvin algorithm. Only needed when d1 = 'robust_mahalanobis'.
            weights: the sample weights. Only used if provided and d1 = 'robust_mahalanobis'.  
            fast_VG: whether the geometric variability estimation will be full (False) or fast (True).
            VG_sample_size: sample size to be used to make the estimation of the geometric variability.
            VG_n_samples: number of samples to be used to make the estimation of the geometric variability.
            random_state: the random seed used for the (random) sample elements.
        """
        self.p1 = p1 ; self.p2 = p2 ; self.p3 = p3
        self.d1 = d1 ; self.d2 = d2 ; self.d3 = d3
        self.q = q ; self.robust_method = robust_method ; self.alpha = alpha ; 
        self.epsilon = epsilon ; self.n_iters = n_iters
        self.VG_sample_size = VG_sample_size; self.VG_n_samples = VG_n_samples
        self.random_state = random_state ; self.fast_VG = fast_VG; self.weights = weights

    def compute(self, X):
        """
        Compute method.
        
        Parameters:
            X: a pandas/polars data-frame or a numpy array. Represents a data matrix.
            
        Returns:
            D: the Generalized Gower matrix for the data matrix `X`.
        """

        D1, D2, D3 = get_dist_matrices(X=X, p1=self.p1, p2=self.p2, p3=self.p3, 
                                               d1=self.d1, d2=self.d2, d3=self.d3, 
                                               q=self.q, robust_method=self.robust_method, epsilon=self.epsilon, 
                                               alpha=self.alpha, n_iters=self.n_iters, weights=self.weights)
     
        D1_2 = D1**2  ; D2_2 = D2**2 ; D3_2 = D3**2

        if self.fast_VG == True:   
            VG1, VG2, VG3 = vg_ggower_fast_estimation(X=X, p1=self.p1, p2=self.p2, p3=self.p3, 
                                                   d1=self.d1, d2=self.d2, d3=self.d3, 
                                                   robust_method=self.robust_method, alpha=self.alpha,
                                                   VG_sample_size=self.VG_sample_size, VG_n_samples=self.VG_n_samples, 
                                                   random_state=self.random_state, weights=self.weights)
        else:
            VG1, VG2, VG3 = vg(D1_2), vg(D2_2), vg(D3_2)

        D1_std = D1_2/VG1 if VG1 > 0 else D1_2 
        D2_std = D2_2/VG2 if VG2 > 0 else D2_2 
        D3_std = D3_2/VG3 if VG3 > 0 else D3_2
        D_2 = D1_std + D2_std + D3_std
        D = np.sqrt(D_2)

        return D 

################################################################################
    
class GGowerDist: 
    """
    Calculates the Generalized Gower distance for a pair of data observations.
    """

    def __init__(self, p1, p2, p3, d1='euclidean', d2='sokal', d3='matching', q=1, robust_method='trimmed', alpha=0.05, epsilon=0.05, n_iters=20,
                 VG_sample_size=300, VG_n_samples=5, random_state=123, weights=None):
        """
        Constructor method.
        
        Parameters:
            p1, p2, p3: number of quantitative, binary and multi-class variables in the considered data matrix, respectively. Must be a non negative integer.
            d1: name of the distance to be computed for quantitative variables. Must be an string in ['euclidean', 'minkowski', 'canberra', 'mahalanobis', 'robust_mahalanobis']. 
            d2: name of the distance to be computed for binary variables. Must be an string in ['sokal', 'jaccard'].
            d3: name of the distance to be computed for multi-class variables. Must be an string in ['matching'].
            q: the parameter that defines the Minkowski distance. Must be a positive integer.
            robust_method: the robust_method to be used for computing the robust covariance matrix. Only needed when d1 = 'robust_mahalanobis'.
            epsilon: parameter used by the Delvin algorithm that is used when computing the robust covariance matrix. Only needed when d1 = 'robust_mahalanobis'.
            n_iter: maximum number of iterations used by the Delvin algorithm. Only needed when d1 = 'robust_mahalanobis'.
            weights: the sample weights. Only used if provided and d1 = 'robust_mahalanobis'.  
            VG_sample_size: sample size to be used to make the estimation of the geometric variability.
            VG_n_samples: number of samples to be used to make the estimation of the geometric variability.
            random_state: the random seed used for the (random) sample elements.
        """
        self.p1 = p1 ; self.p2 = p2 ; self.p3 = p3
        self.d1 = d1 ; self.d2 = d2 ; self.d3 = d3
        self.q = q ; self.robust_method = robust_method ; self.alpha = alpha ; 
        self.epsilon = epsilon ; self.n_iters = n_iters
        self.VG_sample_size = VG_sample_size; self.VG_n_samples = VG_n_samples
        self.random_state = random_state; self.weights = weights

    def fit(self, X) :
        """
        Fit method that computes the geometric variability and covariance matrix to be used in 'compute' method, if needed.
        
        Parameters:
            X: a pandas/polars data-frame or a numpy array. Represents a data matrix.
            
        Returns:
            D: the Generalized Gower matrix for the data matrix `X`.
        """
        p1 = self.p1 ; p2 = self.p2 ; p3 = self.p3
        d1 = self.d1 ; d2 = self.d2 ; d3 = self.d3
        self.S, self.S_robust = None, None

        if d1 in ['robust_mahalanobis', 'mahalanobis']:

            if isinstance(X, (pl.DataFrame, pd.DataFrame)) :
                X = X.to_numpy()
                
            X_quant = X[:, 0:p1] 

            if d1 == 'robust_mahalanobis':
                self.S_robust = S_robust(X=X_quant, method=self.robust_method, alpha=self.alpha, 
                                            epsilon=self.epsilon, n_iters=self.n_iters, weights=self.weights)
            elif d1 == 'mahalanobis':
                self.S = np.cov(X_quant, rowvar=False)

        self.VG1, self.VG2, self.VG3 = vg_ggower_fast_estimation(X=X, p1=p1, p2=p2, p3=p3, d1=d1, d2=d2, d3=d3, robust_method=self.robust_method, 
                                                                 alpha=self.alpha, epsilon=self.epsilon, n_iters=self.n_iters,
                                                                 VG_sample_size=self.VG_sample_size, VG_n_samples=self.VG_n_samples, 
                                                                 random_state=self.random_state, weights=self.weights)
    
    def compute(self, xi, xr):
        """
        Compute method.
        
        Parameters:
            xi, xr: a pair of quantitative vectors. They represent a couple of statistical observations.
            
        Returns:
            dist: the Generalized Gower distance between the observations `xi` and `xr`.
        """
        dist1, dist2, dist3 = get_distances(xi=xi, xr=xr, p1=self.p1, p2=self.p2, p3=self.p3, 
                                            d1=self.d1, d2=self.d2, d3=self.d3, 
                                            q=self.q, S=self.S, S_robust=self.S_robust)
        
        dist1_2 = dist1**2 ; dist2_2 = dist2**2 ; dist3_2 = dist3**2
        dist1_2_std = dist1_2/self.VG1 if self.VG1 > 0 else dist1_2 
        dist2_2_std = dist2_2/self.VG2 if self.VG2 > 0 else dist2_2 
        dist3_2_std = dist3_2/self.VG3 if self.VG3 > 0 else dist3_2 
        dist_2 = dist1_2_std + dist2_2_std + dist3_2_std
        dist = np.sqrt(dist_2)

        return dist

################################################################################

def ggower_dist(xi, xr, p1, p2, p3, d1='euclidean', d2='sokal', d3='matching', 
                q=1, S=None, S_robust=None, VG1=None, VG2=None, VG3=None):
   
    dist1, dist2, dist3 = get_distances(xi=xi, xr=xr, p1=p1, p2=p2, p3=p3, 
                                        d1=d1, d2=d2, d3=d3, 
                                        q=q, S=S, S_robust=S_robust)
        
    dist1_2 = dist1**2 ; dist2_2 = dist2**2 ; dist3_2 = dist3**2
    dist1_2_std = dist1_2/VG1 if VG1 > 0 else dist1_2 
    dist2_2_std = dist2_2/VG2 if VG2 > 0 else dist2_2 
    dist3_2_std = dist3_2/VG3 if VG3 > 0 else dist3_2 
    dist_2 = dist1_2_std + dist2_2_std + dist3_2_std
    dist = np.sqrt(dist_2)

    return dist

################################################################################
    
def simple_gower_dist(xi, xr, X, p1, p2, p3) :
    """
    Compute method.
    
    Parameters:
        xi, xr: a pair of quantitative vectors. They represent a couple of statistical observations.
        X: a pandas/polars data-frame or a numpy array. It represents a data matrix.
        p1, p2, p3: number of quantitative, binary and multi-class variables in the considered data matrix, respectively. Must be a non negative integer.

    Returns:
        dist: the Simple Gower distance between the observations `xi` and `xr`.
    """    

    if isinstance(X, (pl.DataFrame, pd.DataFrame)) :
        X = X.to_numpy()  
    if isinstance(xi, (pl.DataFrame, pd.DataFrame)) :
        xi = xi.to_numpy().flatten()
    elif isinstance(xi, (pd.Series, pl.Series)) :
        xi = xi.to_numpy() 
    if isinstance(xr, (pl.DataFrame, pd.DataFrame)) :
        xr = xr.to_numpy().flatten()
    elif isinstance(xi, (pd.Series, pl.Series)) :
        xr = xr.to_numpy() 

    dist = get_dist_functions()

    X_quant = X[:,0:p1]  
    xi_quant = xi[0:p1] ; xr_quant = xr[0:p1] ; 
    xi_bin = xi[(p1):(p1+p2)] ; xr_bin = xr[(p1):(p1+p2)]
    xi_multi = xi[(p1+p2):(p1+p2+p3)] ; xr_multi = xr[(p1+p2):(p1+p2+p3)]
    R = np.max(X_quant, axis=0) - np.min(X_quant, axis=0)

    dist1 = np.sum(np.abs(xi_quant - xr_quant)/R) if p1 > 0 else 0
    dist2 = dist['jaccard'](xi_bin, xr_bin) if p2 > 0 else 0
    dist3 = dist['hamming'](xi_multi, xr_multi) if p3 > 0 else 0
    dist = dist1 + dist2 + dist3

    return dist

################################################################################

'''
def simple_gower_dist_matrix(X, p1, p2, p3):

    if isinstance(X, (pl.DataFrame, pd.DataFrame)) :
        X = X.to_numpy()  

    D = np.zeros((len(X), len(X)))

    for i in range(len(X)):
        for r in range(len(X)):
            if i <= r:
                D[i,r] = simple_gower_dist(xi=X[i,:], xr=X[r,:], X=X, 
                                           p1=p1, p2=p2, p3=p3)

    D = D + D.T - np.diag(D.diagonal())

    return D
'''

def simple_gower_dist_matrix(X, p1, p2, p3):
    """
    Cálculo matricial de la distancia simple de Gower entre todas las filas de X.

    Parameters:
        X: np.ndarray o DataFrame (se convierte a np.ndarray).
        p1: número de columnas numéricas.
        p2: número de columnas binarias.
        p3: número de columnas categóricas (multi-clase).

    Returns:
        D: matriz de distancias (n x n) con la distancia de Gower simple entre observaciones.
    """

    # Convertir DataFrame si fuera necesario
    if isinstance(X, (pd.DataFrame, pl.DataFrame)):
        X = X.to_numpy()

    dist_matrix = get_dist_matrix_functions()

    # Separar bloques
    X_quant = X[:, 0:p1] if p1 > 0 else None
    X_bin = X[:, p1:p1 + p2] if p2 > 0 else None
    X_multi = X[:, p1 + p2:p1 + p2 + p3] if p3 > 0 else None

    n = X.shape[0]
    D = np.zeros((n, n))

    # Distancia cuantitativa: Manhattan normalizada por rango
    if p1 > 0:
        R = np.max(X_quant, axis=0) - np.min(X_quant, axis=0)
        R[R == 0] = 1  # evitar división por cero
        X_quant_norm = X_quant / R
        dist_quant = dist_matrix['minkowski'](X_quant_norm, q=1)
        D += dist_quant

    # Distancia binaria: Jaccard
    if p2 > 0:
        dist_bin = dist_matrix['jaccard'](X_bin)
        D += dist_bin

    # Distancia categórica: Hamming (simple coincidencia)
    if p3 > 0:
        dist_multi = dist_matrix['hamming'](X_multi)
        D += dist_multi

    return D


################################################################################
    
class RelMSDistMatrix: 
    """
    Calculates the Related Metric Scaling matrix for a data matrix.
    """

    def __init__(self, p1,p2,p3,d1='euclidean',d2='sokal',d3='matching',q=1, robust_method='trimmed', 
                 epsilon=0.05, alpha=0.05, n_iters=20, weights=None, 
                 fast_VG=False, VG_sample_size=300, VG_n_samples=5, random_state=123):
        """
        Constructor method.
        
        Parameters:
            p1, p2, p3: number of quantitative, binary and multi-class variables in the considered data matrix, respectively. Must be a non negative integer.
            d1: name of the distance to be computed for quantitative variables. Must be an string in ['euclidean', 'minkowski', 'canberra', 'mahalanobis', 'robust_mahalanobis']. 
            d2: name of the distance to be computed for binary variables. Must be an string in ['sokal', 'jaccard'].
            d3: name of the distance to be computed for multi-class variables. Must be an string in ['matching'].
            q: the parameter that defines the Minkowski distance. Must be a positive integer.
            robust_method: the robust_method to be used for computing the robust covariance matrix. Only needed when d1 = 'robust_mahalanobis'.
            epsilon: parameter used by the Delvin algorithm that is used when computing the robust covariance matrix. Only needed when d1 = 'robust_mahalanobis'.
            n_iters: maximum number of iterations used by the Delvin algorithm. Only needed when d1 = 'robust_mahalanobis'.
            weights: the sample weights. Only used if provided and d1 = 'robust_mahalanobis'.  
        """
        self.p1 = p1 ; self.p2 = p2 ; self.p3 = p3
        self.d1 = d1 ; self.d2 = d2 ; self.d3 = d3
        self.q = q ; self.robust_method = robust_method ; self.alpha = alpha ; self.fast_VG = fast_VG;
        self.VG_sample_size = VG_sample_size; self.VG_n_samples = VG_n_samples; self.random_state = random_state;
        self.epsilon = epsilon ; self.n_iters = n_iters ; self.weights = weights


    def compute(self, X, tol=1e-6, d=2.5, Gs_PSD_transformation=True):
        """
        Compute method.
        
        Parameters:
            X: a pandas/polars data-frame or a numpy array. Represents a data matrix.
            tol: a tolerance value to round the close-to-zero eigenvalues of the Gramm matrices.
            Gs_PSD_trans: controls if a transformation is applied to enforce positive semi-definite Gramm matrices.
            d: a parameter that controls the omega definition involved in the transformation mentioned above.
            
        Returns:
            D: the Related Metric Scaling matrix for the data matrix `X`.
        """
        D1, D2, D3  = get_dist_matrices(X=X, p1=self.p1, p2=self.p2, p3=self.p3, 
                                               d1=self.d1, d2=self.d2, d3=self.d3, 
                                               q=self.q, robust_method=self.robust_method, epsilon=self.epsilon, 
                                               alpha=self.alpha, n_iters=self.n_iters, weights=self.weights)
       
        D1_2 = D1**2  ; D2_2 = D2**2 ; D3_2 = D3**2

        if self.fast_VG == True:   
            VG1, VG2, VG3 = vg_ggower_fast_estimation(X=X, p1=self.p1, p2=self.p2, p3=self.p3, 
                                                   d1=self.d1, d2=self.d2, d3=self.d3, 
                                                   robust_method=self.robust_method, alpha=self.alpha,
                                                   VG_sample_size=self.VG_sample_size, VG_n_samples=self.VG_n_samples, 
                                                   random_state=self.random_state, weights=self.weights)
        else:
            VG1, VG2, VG3 = vg(D1_2), vg(D2_2), vg(D3_2)

        D1_std = D1_2/VG1 if VG1 > 0 else D1_2 
        D2_std = D2_2/VG2 if VG2 > 0 else D2_2 
        D3_std = D3_2/VG3 if VG3 > 0 else D3_2
 
        n = len(D1)
        ones = np.ones((n, 1)) 
        ones_T = np.ones((1, n))
        ones_M = np.ones((n, n))
        I = np.identity(n)
        H = I - (1/n)*(ones @ ones_T)
        G_1 = -(1/2)*(H @ D1_std @ H)
        G_2 = -(1/2)*(H @ D2_std @ H) 
        G_3 = -(1/2)*(H @ D3_std @ H)

        if Gs_PSD_transformation == True :

            v1 = np.real(np.linalg.eigvals(G_1))
            v2 = np.real(np.linalg.eigvals(G_2))
            v3 = np.real(np.linalg.eigvals(G_3))
            v1[np.isclose(v1, 0, atol=tol)] = 0 
            v2[np.isclose(v2, 0, atol=tol)] = 0 
            v3[np.isclose(v3, 0, atol=tol)] = 0
            G1_PSD = np.all(v1 >= 0)
            G2_PSD = np.all(v2 >= 0) 
            G3_PSD = np.all(v3 >= 0)

            if not G1_PSD :
                
                print('G1 is not PSD, a transformation to force it will be applied.')

                omega = d * np.abs(np.min(v1))  
                D1_std  = D1_std + omega*ones_M - omega*I
                G_1 = -(1/2)*(H @ D1_std @ H)

            if not G2_PSD :

                print('G2 is not PSD, a transformation to force it will be applied.')
                omega = d * np.abs(np.min(v2)) 
                D2_std = D2_std + omega*ones_M - omega*I
                G_2 = -(1/2)*(H @ D2_std @ H)

            if not G3_PSD :

                print('G3 is not PSD, a transformation to force it will be applied.')
                omega = d * np.abs(np.min(v3))  
                D3_std = D3_std + omega*ones_M - omega*I
                G_3 = -(1/2)*(H @ D3_std @ H) 
        
        U1, S1, V1 = np.linalg.svd(G_1) 
        U2, S2, V2 = np.linalg.svd(G_2)   
        U3, S3, V3 = np.linalg.svd(G_3)
        S1 = np.clip(S1, 0, None)
        S2 = np.clip(S2, 0, None)
        S3 = np.clip(S3, 0, None)
        sqrtG1 = U1 @ np.diag(np.sqrt(S1)) @ V1 
        sqrtG2 = U2 @ np.diag(np.sqrt(S2)) @ V2 
        sqrtG3 = U3 @ np.diag(np.sqrt(S3)) @ V3

        G = G_1 + G_2 + G_3 - (1/3)*(sqrtG1@sqrtG2 + sqrtG1@sqrtG3 + sqrtG2@sqrtG1 + sqrtG2@sqrtG3 + sqrtG3@sqrtG1 + sqrtG3@sqrtG2)
        g = np.diag(G) 
        g =  np.reshape(g, (len(g), 1))  
        g_T = np.reshape(g, (1, len(g)))   
        D_2_ = g @ ones_T + ones @ g_T - 2*G
        D_2_[np.isclose(D_2_, 0, atol=tol)] = 0
        D = np.sqrt(D_2_)
 
        return D    

################################################################################

def data_preprocessing(X, frac_sample_size, random_state):
    """
    Preprocess data in the way as needed by `FastGG` class.

    Parameters (inputs)
    ----------
    X: a pandas/polars data-frame.
    frac_sample_size: the sample size in proportional terms.
    random_state: the random seed for the random elements of the function.

    Returns (outputs)
    -------
    X_sample: a polars df with the sample of `X`.
    X_out_sample: a polars df with the out of sample of `X`.
    sample_index: the index of the sample observations/rows.
    out_sample_index: the index of the out of sample observations/rows.
    """

    if not (0 < frac_sample_size <= 1):
       raise ValueError('frac_sample_size must be in (0,1].')

    if isinstance(X, (pd.DataFrame, pl.DataFrame)):
        X = X.to_numpy()
    
    n = len(X)

    if frac_sample_size < 1:
        n_sample = int(frac_sample_size*n)
        index = np.arange(0,n)
        np.random.seed(random_state)
        sample_index = np.random.choice(index, size=n_sample, replace=False)
        out_sample_index = np.array([x for x in index if x not in sample_index])
        X_sample = X[sample_index,:] 
        X_out_sample = X[out_sample_index,:] 
    else:
        X_sample = X
        sample_index =  np.arange(0,n)
        X_out_sample = np.array([])
        out_sample_index = np.array([])

    return X_sample, X_out_sample, sample_index, out_sample_index

################################################################################

class FastGGowerDistMatrix:
    """
    Calculates the the Generalized Gower matrix of a sample of a given data matrix.
    """

    def __init__(self, frac_sample_size=0.1, random_state=123, p1=None, p2=None, p3=None, 
                 d1='robust_mahalanobis', d2='jaccard', d3='matching', 
                 robust_method='trimmed', alpha=0.05, epsilon=0.05, n_iters=20, q=1, 
                 fast_VG=False, VG_sample_size=1000, VG_n_samples=5, weights=None) :
        """
        Constructor method.
        
        Parameters:
            frac_sample_size: the sample size in proportional terms.
            p1, p2, p3: number of quantitative, binary and multi-class variables in the considered data matrix, respectively. Must be a non negative integer.
            d1: name of the distance to be computed for quantitative variables. Must be an string in ['euclidean', 'minkowski', 'canberra', 'mahalanobis', 'robust_mahalanobis']. 
            d2: name of the distance to be computed for binary variables. Must be an string in ['sokal', 'jaccard'].
            d3: name of the distance to be computed for multi-class variables. Must be an string in ['matching'].
            q: the parameter that defines the Minkowski distance. Must be a positive integer.
            robust_method: the method to be used for computing the robust covariance matrix. Only needed when d1 = 'robust_mahalanobis'.
            alpha : a real number in [0,1] that is used if `method` is 'trimmed' or 'winsorized'. Only needed when d1 = 'robust_mahalanobis'.
            epsilon: parameter used by the Delvin algorithm that is used when computing the robust covariance matrix. Only needed when d1 = 'robust_mahalanobis'.
            n_iters: maximum number of iterations used by the Delvin algorithm. Only needed when d1 = 'robust_mahalanobis'.
            fast_VG: whether the geometric variability estimation will be full (False) or fast (True).
            VG_sample_size: sample size to be used to make the estimation of the geometric variability.
            VG_n_samples: number of samples to be used to make the estimation of the geometric variability.
            random_state: the random seed used for the (random) sample elements.
            weights: the sample weights. Only used if provided and d1 = 'robust_mahalanobis'.  
        """
        self.random_state = random_state; self.frac_sample_size = frac_sample_size; self.p1 = p1; self.p2 = p2; self.p3 = p3; 
        self.d1 = d1; self.d2 = d2; self.d3 = d3; self.robust_method = robust_method; self.alpha = alpha; self.epsilon = epsilon; 
        self.n_iters = n_iters; self.fast_VG = fast_VG; self.VG_sample_size = VG_sample_size; self.VG_n_samples = VG_n_samples; 
        self.q = q; self.weights = weights

    def compute(self, X):
        """
        Compute method: computes the Generalized Gower function for the defined sample of data.
        
        Parameters:
            X: a pandas/polars data-frame or a numpy array. Represents a data matrix.
        """

        X_sample, X_out_sample, sample_index, out_sample_index = data_preprocessing(X=X, frac_sample_size=self.frac_sample_size, 
                                                                                    random_state=self.random_state)
       
        sample_weights = self.weights[sample_index] if self.weights is not None else None

        GGower_matrix = GGowerDistMatrix(p1=self.p1, p2=self.p2, p3=self.p3, 
                                         d1=self.d1, d2=self.d2, d3=self.d3, q=self.q,
                                         robust_method=self.robust_method, alpha=self.alpha, 
                                         epsilon=self.epsilon, n_iters=self.n_iters,
                                         fast_VG=self.fast_VG, VG_sample_size=self.VG_sample_size, 
                                         VG_n_samples=self.VG_n_samples, weights=sample_weights)
        
        self.D_GGower = GGower_matrix.compute(X=X_sample)
        self.sample_index = sample_index
        self.out_sample_index = out_sample_index
        self.X_sample = X_sample
        self.X_out_sample = X_out_sample

################################################################################


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="robust-mixed-dist",
    version="0.1.5",
    author="Fabio Scielzo Ortiz",
    author_email="fabio.scielzoortiz@gmail.com",
    description="Compute statistical robust distances for mixed data.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/FabioScielzoOrtiz/robust_mixed_dist-package",  # add your project URL here
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=['polars', 'numpy', 'pandas', 'scipy'],
    python_requires=">=3.7"
)
